/* #includes */ /*{{{C}}}*//*{{{*/
#include "config.h"

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#ifdef HAVE_GETTEXT
#include <libintl.h>
#define _(String) gettext(String)
#else
#define _(String) String
#endif
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "getopt.h"
/*}}}*/
/* #defines */ /*{{{*/
#define ISALPHA(c) ((c>='a' && c<='z') || (c>='A' && c<='Z'))
/*}}}*/

/* types */ /*{{{*/
struct Url
{
  char *url;
  int number;
  struct Url *next;
};
/*}}}*/
/* variables */ /*{{{*/
static int intitle,inlist,inheader,inpre,inword,inwhite;
static int words=0,skipheaders=0,skiplists=0,pretty=0;
static int first=1;
static const char *file;
static int line;
/*}}}*/

static int mystrcasecmp(const char *s, const char *t) /*{{{*/
{
  int x;

  while ((x=tolower(*s)-tolower(*t))==0 && *s) { ++s; ++t; }
  return x;
}
/*}}}*/
static int condputchar(int c) /*{{{*/
{
  static int nls=2;
  static int lastc='\n';

  if (words || ((!skiplists || !inlist) && (!skipheaders || (!inheader && !intitle))))
  {
    if (pretty)
    {
      if (c=='\n')
      {
        if (++nls>2) return c;
      }
      else
      {
        nls=0;
      }
      if (c=='\n') ++line;
      lastc=c;
      return putchar(c);
    }
    else return putchar(c);
  }
  else if (!pretty && c=='\n') return putchar(c);
}
/*}}}*/
static void wordputchar(int c) /*{{{*/
{
  if (words)
  {
    if (isalpha(c) || c=='_') { inword=1; inwhite=0; condputchar(c); }
    else if (inword && !inwhite) { inword=0; inwhite=1; condputchar('\n'); }
  }
  else condputchar(c);
}
/*}}}*/
static void dehtml(FILE *fp, const char *fileName) /*{{{*/
{
  int c;
  char href[512];
  struct Url *urls,**lasturl;

  line=1;
  file=fileName;
  intitle=inlist=inheader=inpre=0;
  href[0]='\0';
  urls=(struct Url*)0;
  lasturl=&urls;
  while ((c=getc(fp))!=EOF)
  {
    if (c=='<') /* tag */ /*{{{*/
    {
      char tag[sizeof("/address")];
      char attribute[sizeof("href")];
      int i;

      if (inword) inword=0;
      /* tag name */ /*{{{*/
      i=0;
      while ((c=getc(fp))!=EOF && c!='>' && c!=' ' && c!='\n')
      {
        if (i<sizeof(tag)-1) tag[i++]=tolower(c);
      }
      tag[i]='\0';
      if (c=='\n')
      {
        ++line;
        condputchar('\n');
      }
      if (i && i<sizeof(tag))
      {
        if (mystrcasecmp(tag,"p")==0 || mystrcasecmp(tag,"hr")==0) /*{{{*/
        {
          if (!words && pretty) { condputchar('\n'); condputchar('\n'); }
        }
        /*}}}*/
        else if (mystrcasecmp(tag,"br")==0) /*{{{*/
        {
          if (!words && pretty) condputchar('\n');
        }
        /*}}}*/
        else if (mystrcasecmp(tag,"title")==0) intitle=1;
        else if (mystrcasecmp(tag,"/title")==0) intitle=0;
        else if (tolower(tag[0])=='h' && isdigit(tag[1]) && tag[2]=='\0') /*{{{*/
        {
          if (!words && pretty)
          {
            condputchar('\n');
            condputchar('\n');
          }
          ++inheader;
        }
        /*}}}*/
        else if (tag[0]=='/' && tolower(tag[1])=='h' && isdigit(tag[2]) && tag[3]=='\0') /*{{{*/
        {
          if (!words && pretty)
          {
            condputchar('\n');
            condputchar('\n');
          }
          if (inheader) --inheader;
        }
        /*}}}*/
        else if (mystrcasecmp(tag,"pre")==0) inpre=1;
        else if (mystrcasecmp(tag,"/pre")==0) inpre=0;
        else if (mystrcasecmp(tag,"dl")==0) ++inlist;
        else if (mystrcasecmp(tag,"/dl")==0) { if (inlist) --inlist; }
        else if (mystrcasecmp(tag,"ul")==0) ++inlist;
        else if (mystrcasecmp(tag,"/ul")==0) { if (inlist) --inlist; }
        else if (mystrcasecmp(tag,"ol")==0) ++inlist;
        else if (mystrcasecmp(tag,"/ol")==0) { if (inlist) --inlist; }
        else if (mystrcasecmp(tag,"/a")==0 && href[0]) /*{{{*/
        {
          struct Url *u;
          char n[32],*s;
          int number=0;

          for (u=urls; u && strcmp(u->url,href); u=u->next) number=u->number;
          if (u==(struct Url*)0)
          {
            u=malloc(sizeof(struct Url));
            u->number=number+1;
            u->url=strcpy(malloc(strlen(href)+1),href);
            u->next=(struct Url*)0;
            *lasturl=u;
            lasturl=&u->next;
          }
          snprintf(n,sizeof(n)," [%d]",u->number);
          for (s=n; *s; ++s) wordputchar(*s);
          href[0]='\0';
        }
        /*}}}*/
      }
      /*}}}*/
      if (c!=EOF && c!='>') /* tag attributes */ /*{{{*/
      {
        enum { EMPTY, ATTRIBUTE, EQ, VALUE, QUOTEDVALUE } state=EMPTY;
        int output_value=0;
        int a_href=0;

        do
        {
          c=getc(fp);
          if (c=='\n')
          {
            ++line;
            condputchar('\n');
          }
          switch (state)
          {
            case EMPTY: /*{{{*/
            {
              if (ISALPHA(c))
              {
                state=ATTRIBUTE;
                i=0;
                attribute[i++]=c;
              }
              break;
            }
            /*}}}*/
            case ATTRIBUTE: /*{{{*/
            {
              if (ISALPHA(c))
              {
                if (i<sizeof(attribute)-1) attribute[i++]=tolower(c);
              }
              else
              {
                attribute[i]='\0';
                if (c=='=')
                {
                  state=EQ;
                  a_href=(strcmp(tag,"a")==0) && (strcmp(attribute,"href")==0);
                  output_value=(strcmp(tag,"img")==0) && (strcmp(attribute,"alt")==0);
                }
                else state=EMPTY;
              }
              break;
            }
            /*}}}*/
            case EQ: /*{{{*/
            {
              i=0;
              if (c=='"') state=QUOTEDVALUE;
              else
              {
                state=VALUE;
                if (output_value) wordputchar(c);
              }
              break;
            }
            /*}}}*/
            case QUOTEDVALUE: /*{{{*/
            {
              if (c=='"')
              {
                if (a_href)
                {
                  href[i]='\0';
                  a_href=0;
                }
                output_value=0;
                state=EMPTY;
              }
              else if (a_href)
              {
                if (i<sizeof(href)-1) href[i++]=c;
              }
              else if (output_value) condputchar(c);
              break;
            }
            /*}}}*/
            case VALUE: /*{{{*/
            {
              if (c==' ')
              {
                if (a_href)
                {
                  a_href=0;
                  href[i]='\0';
                }
                output_value=0;
                state=EMPTY;
              }
              else if (a_href)
              {
                if (i<sizeof(href)-1) href[i++]=c;
              }
              else if (output_value) wordputchar(c);
              break;
            }
            /*}}}*/
          }
        } while (c!=EOF && c!='>');
      }
      /*}}}*/
    }
    /*}}}*/
    else if (c=='&') /* entity */ /*{{{*/
    {
      char entity[73];
      int i=0;

      if ((c=getc(fp))=='#')
      {
        c=getc(fp);
        if (isdigit(c))
        {
          int numeric=c-'0';

          while ((c=getc(fp))!=EOF && isdigit(c))
          {
            numeric=numeric*10+(c-'0');
          }
          wordputchar(numeric);
          if (c!=';') wordputchar(c);
        }
        else
        {
          wordputchar('&');
          wordputchar('#');
        }
      }
      else if (ISALPHA(c) || isdigit(c) || c=='.' || c=='-')
      {
        /* variables */ /*{{{*/
          struct
          {
            const char *name;
            char value;
          }
          const *eptr,
          entities[]=
          {
            { "gt",     '>' },
            { "lt",     '<' },
            { "amp",    '&' },
            { "quot",   '"' },
            { "AElig",  '' },
            { "Aacute", '' },
            { "Acirc",  '' },
            { "Agrave", '' },
            { "Aring",  '' },
            { "Atilde", '' },
            { "Auml",   '' },
            { "Ccedil", '' },
            { "ETH",    '' },
            { "Eacute", '' },
            { "Ecirc",  '' },
            { "Egrave", '' },
            { "Euml",   '' },
            { "Iacute", '' },
            { "Icirc",  '' },
            { "Igrave", '' },
            { "Iuml",   '' },
            { "Ntilde", '' },
            { "Oacute", '' },
            { "Ocirc",  '' },
            { "Ograve", '' },
            { "Oslash", '' },
            { "Otilde", '' },
            { "Ouml",   '' },
            { "THORN",  '' },
            { "Uacute", '' },
            { "Ucirc",  '' },
            { "Ugrave", '' },
            { "Uuml",   '' },
            { "Yacute", '' },
            { "aacute", '' },
            { "acirc",  '' },
            { "aelig",  '' },
            { "agrave", '' },
            { "aring",  '' },
            { "atilde", '' },
            { "auml",   '' },
            { "ccedil", '' },
            { "eacute", '' },
            { "ecirc",  '' },
            { "egrave", '' },
            { "eth",    '' },
            { "euml",   '' },
            { "iacute", '' },
            { "icirc",  '' },
            { "igrave", '' },
            { "iuml",   '' },
            { "nbsp",   '' },
            { "ntilde", '' },
            { "oacute", '' },
            { "ocirc",  '' },
            { "ograve", '' },
            { "oslash", '' },
            { "otilde", '' },
            { "ouml",   '' },
            { "szlig",  '' },
            { "thorn",  '' },
            { "uacute", '' },
            { "ucirc",  '' },
            { "ugrave", '' },
            { "uuml",   '' },
            { "yacute", '' },
            { "yuml",   '' }
          };
          /*}}}*/

        entity[i++]=c;
        while ((c=getc(fp))!=EOF && (ISALPHA(c) || isdigit(c) || c=='.' || c=='-'))
        {
          if (i<sizeof(entity)-1) entity[i++]=c;
        }
        entity[i]='\0';
        for (eptr=entities; eptr<entities+sizeof(entities)/sizeof(entities[0]); ++eptr)
        {
          if (strcmp(eptr->name,entity)==0)
          {
            wordputchar(eptr->value);
            if (c!=';') wordputchar(c);
            goto continueLoop;
          }
          else if (strcmp(entity,"hellip")==0)
          {
            wordputchar('.');
            wordputchar('.');
            wordputchar('.');
            goto continueLoop;
          }
        }
        wordputchar('&');
        for (i=0; entity[i]; ++i) wordputchar(entity[i]);
        wordputchar(c);
      }
      else
      {
        wordputchar('&');
        wordputchar(c);
      }
    }
    /*}}}*/
    else if (c=='\n') /* new line */ /*{{{*/
    {
      ++line;
      wordputchar(c);
    }
    /*}}}*/
    else wordputchar(c);
    continueLoop:;
  }
  wordputchar('\n');
  while (urls)
  {
    char n[32],*s;
    struct Url *f;

    snprintf(n,sizeof(n),"[%d] ",urls->number);
    for (s=n; *s; ++s) wordputchar(*s);
    for (s=urls->url; *s; ++s) wordputchar(*s);
    wordputchar('\n');
    free(urls->url);
    f=urls;
    urls=urls->next;
    free(f);
  }
}
/*}}}*/

int main(int argc, char *argv[]) /*{{{*/
{
  /* variable declarations */ /*{{{*/
  FILE *in;
  int usage=0;
  int c;
  static struct option lopts[]=
  {
    { "word-list", no_argument, 0, 'w' },
    { "skip-headers", no_argument, 0, 's' },
    { "skip-lists", no_argument, 0, 'l' },
    { "prettyprint", no_argument, 0, 'p' },
    { "urls", no_argument, 0, 'u' },
    { "help", no_argument, 0, 'h' },
    { "version", no_argument, 0, 'v' },
    { (const char*)0, 0, 0, '\0' }
  };
  /*}}}*/

  setlocale(LC_MESSAGES,"");
  setlocale(LC_CTYPE,"");
#ifdef HAVE_GETTEXT
  bindtextdomain("dehtml",LOCALEDIR);
  textdomain("dehtml");
#endif
  /* parse arguments */ /*{{{*/
  while ((c=getopt_long(argc,argv,"wslp?h",lopts,(int*)0))!=EOF) switch(c)
  {
    case 'w': words=1; break;
    case 's': skipheaders=1; break;
    case 'l': skiplists=1; break;
    case 'p': pretty=1; break;
    case 'h': usage=2; break;
    case 'v': printf("dehtml " VERSION "\n"); exit(0);
    default: usage=1;
  }
  if (usage==1)
  {
    fprintf(stderr,_("Usage: dehtml [-w] [-s] [-l] [-p] [file ...]\n"));
    fprintf(stderr,"\n");
    fprintf(stderr,_("Try `dehtml -h' or `dehtml --help' for more information.\n"));
    exit(1);
  }
  if (usage==2)
  {
    fprintf(stderr,_("Usage: dehtml [-w] [-s] [-l] [-p] [file ...]\n"));
    fprintf(stderr,"\n");
    fprintf(stderr,_("Remove HTML constructs from documents.\n"));
    fprintf(stderr,"\n");
    fprintf(stderr,_("-w, --word-list     output a word list\n"));
    fprintf(stderr,_("-s, --skip-headers  do not output headers\n"));
    fprintf(stderr,_("-l, --skip-lists    do not output lists\n"));
    fprintf(stderr,_("-p, --pretty-print  pretty printed output\n"));
    fprintf(stderr,_("-h, --help          display this help and exit\n"));
    fprintf(stderr,_("    --version       display version and exit\n"));
    fprintf(stderr,"\n");
    fprintf(stderr,_("Report bugs to <michael@moria.de>.\n"));
    exit(0);
  }
  /*}}}*/
  /* dehtml stdin or files, if any */ /*{{{*/
  if (optind<argc) while (optind<argc)
  {
    if ((in=fopen(argv[optind],"r"))==(FILE*)0)
    {
      fprintf(stderr,_("dehtml: Opening `%s' failed (%s).\n"),argv[optind],strerror(errno));
      exit(1);
    }
    dehtml(in,argv[optind]);
    fclose(in);
    ++optind;
  }
  else dehtml(stdin,(const char*)0);
  if (fclose(stdout)==-1)
  {
    fprintf(stderr,_("dehtml: Closing standard output failed (%s).\n"),strerror(errno));
    return 1;
  }
  /*}}}*/
  return 0;
}
/*}}}*/
